/*
 * This is not public software.
 * You should not have access to this code source file.
 * This is public because is not standard in D.
 * D libraries are being completed and we hope that
 * soon some kind of directory package will be available
 * publicly.
 * Please delete this file from your system emmidiatly
 */

module leds.Directory;

private import string;
private import time;
private import linux;

public:
enum DirentTypes
{
	DIRECTORY = 4,
	FILE = 8,
	S_LINK = 10

}

public:
enum StatTypes
{
	DIRECTORY	= 0040000, /* Directory.  */
	CHAR_DEV	= 0020000, /* Character device.  */
	BLOCK_DEV	= 0060000, /* Block device.  */
	FILE		= 0100000, /* Regular file.  */
	FIFO		= 0010000, /* FIFO.  */
	S_LINK		= 0120000, /* Symbolic link.  */
	SOCKET		= 0140000, /* Socket.  */
}


version(linux)
{
	const char[] MAKE_DIR = "mkdir -p ";
}
else
{
	const char[] MAKE_DIR = "md ";
}

private:
extern(C)
{

	struct DIR;
	
	// to create dirs we'll use the mkdir
	
	int system(char *command);

	// to remove a file or dir
	
	int remove(char *pathname);
	
	// this structure will not work on all systems
	// the the size of the integer will vary
	struct Dirent
	{
		uint d_ino;
		uint d_off;
		ushort d_reclen;
		char d_type;
		char[256] d_name;		/* OS limit */
	};

	// directory access

	DIR* opendir(char* name);
	Dirent* readdir(DIR* dir);
	void rewinddir(DIR *dir);

	// stat access

	//int stat(char *file_name, stat *buf);
	// int fstat(int filedes, stat *buf); define in linux.d
	//int lstat(char *file_name, stat *buf);


	// c free

	void free(void*);
}

/**
 * A directory error
 * No error codes for now, just inadequate descriptive message
 */
class DirectoryError : Error
{
	this(char[] message)
	{
		super(message);
	}
}

/**
 * A directory content browser.
 * \todo get meaninfull information on the directory entries
 */
public:
class Directory
{
	int count = 0;

	DIR* dir;
	Dirent* entry;

	Dirent*[] dirs;
	Dirent*[] files;
	

	public:

	bit next()
	{
		if ( entry !== null )
		{
			free(entry);
			entry = null;
		}
		//printf("\n");
		Dirent* entry = readdir(dir);
		if ( entry !== null )
		{
			//printf("entry d_ino    %d\n",entry.d_ino);
			//printf("entry d_off    %d\n",entry.d_off);
			//printf("entry d_reclen %d\n",entry.d_reclen);
			//printf("entry d_type   %c\n",entry.d_type);
			printf("%d %d %.*s\n",count++,(int)entry.d_type,entry.d_name);
		}
		//return entry;
		return entry !== null;
	}

	/**
	 * Resets the read pointer to the first entry
	 */
	void rewind()
	{
		rewinddir(dir);
	}

	private Dirent*[] getEntries(DirentTypes type)
	{
		Dirent*[] entries;
		Dirent* entry;
		do
		{
			entry = readdir(dir);
			if ( entry !== null )
			{
				if ( entry.d_type == type )
				{
					entries ~= entry;
				}
				else
				{
					//free(entry);
				}
			}
		} while (entry !== null);
		return entries;
	}

	/**
	 * get and array with all the subdirectories of this directory
	 * @return and array with all the subdirectories
	 * \todo include the simbolic links
	 * \todo avoid scan the directory twice to get the subdirs and the files
	 */
	Dirent*[] getDirs()
	{
		//freeAndClearDirs();
		rewinddir(dir);
		dirs = getEntries(DirentTypes.DIRECTORY);
		return dirs;
	}

	/**
	 * get and array with all the files of this directory
	 * @return and array with all the files
	 * \todo include the simbolic links
	 * \todo avoid scan the directory twice to get the subdirs and the files
	 */
	Dirent*[] getFiles()
	{
		//freeAndClearDirs();
		rewinddir(dir);
		files = getEntries(DirentTypes.FILE);
		return dirs;
	}

	/**
	 * Gets a sorted array with all the subdirectory names
	 * @return a sorted array with all the subdirectory names
	 */
	char[][] getNames(DirentTypes type)
	{
		char[][] names;
		rewinddir(dir);
		Dirent* entry = readdir(dir);
		int count = 0;
		while ( entry !== null )
		{
			++count;
			if ( entry.d_type == type )
			{
				//printf("Directory adding name %d %.*s\n",count,entry.d_name);
				names[] ~= string.toString(entry.d_name);
			}
			//free(entry);
			entry = readdir(dir);
		}
		return names;
	}

	/**
	 * Gets a sorted array of the subdirectories
	 * @return a sorted array of the subdirectories
	 */
	char[][] getDirNames()
	{
		return getNames(DirentTypes.DIRECTORY);
	}

	/**
	 * Gets a sorted array of the files on this directory
	 * @return a sorted array of the files on this directory
	 */
	char[][] getFileNames()
	{
		return getNames(DirentTypes.FILE);
	}

	/**
	 * Frees the content of the dirs array and sets it's length to 0
	 */
	private void freeAndClearDirs()
	{
		foreach(Dirent* d ; dirs)
		{
			free(d);
		}
		dirs.length = 0;
	}

	/**
	 * Frees the content of the files array and setst it;s length to 0
	 */
	private void freeAndClearFiles()
	{
		foreach(Dirent* f ; files)
		{
			free(f);
		}
		files.length = 0;
	}

	/**
	 * Creates a Directory object for the directory
	 * @param directory must be a directory name
	 */
	this(char[] directory)
	{
		if ( directory === null )
		{
			throw new DirectoryError("Directory name not received");
		}
		//printf("Created new Directory %.*s\n",directory);
		dir = opendir((char*)(directory~"\0"));
		if ( dir === null )
		{
			throw new DirectoryError("can't open directory " ~ directory);
		}
	}

	~this()
	{
		if ( dir !== null )
		{
			//free(dir);
		}
		if ( entry !== null )
		{
			//free(entry);
		}
		//freeAndClearDirs();
		//freeAndClearFiles();
	}

	/**
	 * Creates a new directory 
	 */
	static int mkdir(char[] dir)
	{
		return system(string.toString(MAKE_DIR ~ dir~"\0"));
	}

	public static int removeFile(char[] pathname)
	{
		//printf("Directory.remove pathname = %.*s\n",pathname);
		return remove(string.toString(pathname~"\0"));
	}
	
	
}

///**
// * just testing the thing
// */
//void main(char[][] argv)
//{
//
//	char [] dirName;
//
//	if ( argv.length < 2 )
//	{
//		dirName = ".";
//	}
//	else
//	{
//		dirName = argv[1];
//	}
//
//	Directory dir = new Directory(dirName);
//	bit d;
//	
//	printf("\n###################\n");
//	printf("################### All entries of %.*s\n",dirName);
//	printf("###################\n\n");
//
//	do
//	{
//		d = dir.next();
//	} while(d);
//
//	printf("\n###################\n");
//	printf("################### All subdirectories of %.*s\n",dirName);
//	printf("###################\n\n");
//
//	foreach(Dirent* entry; dir.getDirs() )
//	{
//		printf("\t%.*s\n",entry.d_name);
//	}
//
//	printf("\n###################\n");
//	printf("################### All files of %.*s\n",dirName);
//	printf("###################\n\n");
//
//	foreach(Dirent* entry; dir.getFiles() )
//	{
//		printf("\t%.*s\n",entry.d_name);
//	}
//
//	printf("\n###################\n");
//	printf("################### directory names of %.*s\n",dirName);
//	printf("###################\n\n");
//
//	foreach(char[] name; dir.getDirNames() )
//	{
//		printf("\t%.*s\n",name);
//	}
//
//	printf("\n###################\n");
//	printf("################### file names of %.*s\n",dirName);
//	printf("###################\n\n");
//
//	foreach(char[] name; dir.getFileNames() )
//	{
//		printf("\t%.*s\n",name);
//	}
//
//}
//
