/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.ProjectView;

private import leds.Leds;
private import leds.Directory;
private import leds.Workspace;
private import leds.FileView;
private import leds.Docker;
private import leds.Project;
//private import leds.POpen;
private import leds.ExecConsole;

private import dui.All;
private import string;
private import path;


private class RowType
{
	static const char[] PROJECT = "P";
	static const char[] DIRECTORY = "D";
	static const char[] FILE = "F";
}

extern(C)
{
};

/**
 * Displays the project and their contents.
 * Only one tree exists for all poject on this view:<br>
 * column 1 - the visible string to the user
 * column 2 - the row type { PROJECT, DIRECTORY or FILE }
 * column 3 - the value { always the PATH ? }
 */
public:
class ProjectView : VBox , ButtonClickedListener , Dockable
{
	
	Workspace workspace;
	
	ProjectTreeView treeView;
	ScrolledWindow sw;

	// our tree definition
	/** the columns */
	static GType [3] columns = [DUIType.STRING,DUIType.STRING,DUIType.STRING]; 
	
	/** the tree store */
	ProjectTreeStore projectTreeStore;
	/** the root iter */
	TreeIter iterTop;
	/** the current projects */
	Project [char[]] projects;

	Menu tabMenu;

	Docker docker;

	/**
	 * Creates a new project view for a specific workspace
	 * @param workspace the workspace
	 */
	this(Workspace workspace)
	{
		super(false,0);
		this.workspace = workspace;
		
		sw = new ScrolledWindow(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
		treeView = createProjectTree();
		sw.add(treeView);
		packStart(getToolbar(),false,false,0);
		packStart(sw,true,true,0);
	}

	Widget getWidget()
	{
		return this;
	}
	
	char[] getTabName()
	{
		return "Project";
	}
	
	Menu getTabMenu()
	{
		if ( tabMenu === null )
		{
			tabMenu = new Menu();
			MenuItem item;
			
			//tabMenu.append(new MenuItem(this,"_Close","tabMenu.close"));
			//tabMenu.append(new MenuItem(this,"_Float","tabMenu.float"));
			tabMenu.append(new MenuItem("test"));
		}
		//return tabMenu;
		return null;
	}

	void setDocker(Docker docker)
	{
		this.docker = docker;
	}

	/**
	 * Gets this dockable file view
	 * @return null because with don't have a file view
	 */
	FileView getFileView()
	{
		return null;
	}
	
	/**
	 * detrimines if a close button shoud be visible
	 */
	bit allowCloseButton()
	{
		return false;
	}

	/**
	 * Opens all projects
	 */
	void openActiveProjects()
	{
		openProjects(true);
	}
	
	/**
	 * Opens project depending on the activity flag
	 * @param activeOnly if true only active projects will be open
	 */
	void openProjects(bit activeOnly)
	{
		try
		{
			//printf("openProjects entry %X\n",Leds.leds());
			Directory dir = new Directory(Leds.leds().getProjectsDir());
			foreach(char[]filename ; dir.getFileNames().sort)
			{
				//printf("openProject filename = %.*s\n",filename);
				if ( filename.length>5 && filename[filename.length-4..filename.length] == ".lpj" )
				{
					//printf("openProjects 1\n");
					Project project = new Project(Leds.leds().getProjectFileName(filename));
					project.setWorkspace(workspace);
					//printf("openProjects 2\n");
					//project.dump();
					//if ( project.isActive() )
					//{
						addProject(project);
					//}
					//printf("openProjects 3\n");
				}
			}
		}
		catch ( DirectoryError e)
		{
			//printf("Open projects directory : %.*s\n",e.toString());
		}
		
	}
	
	/**
	 * Adds one project to this view.
	 * \todo don't open same project twice.
	 * \todo don't allow 2 project with the same name
	 */
	void addProject(Project project)
	{

		projects[project.getName()] = project;
		
		//project.dump();
		
		if ( !project.isValid() )
		{
			return;
		}

		static int [3] cols = [0,1,2];
		char [][] vals;
		vals ~= project.getName();
		vals ~= RowType.PROJECT;
		vals ~= project.getBaseDir();
		
		TreeIter projectIter = projectTreeStore.createIter();

		projectTreeStore.set(projectIter,cols,vals);

		try
		{
			loadDir(projectIter, project.getBaseDir(), "", project.getShowHiddenFiles());
			loadFiles(projectIter, project.getBaseDir(),  project.getShowHiddenFiles());
		}
		catch ( DirectoryError e)
		{
			//printf("Add project error : %.*s\n",e.toString());
			TreeIter iterChild = projectTreeStore.append(projectIter);
			char [][] subVals;
			subVals  ~= "not found";
			projectTreeStore.set(iterChild,cols,subVals);
		}

	}

	bit validDir(char[] dir, char[] subdir, bit showHidden)
	{
		//printf("validating dirs %.*s %.*s\n",dir,subdir);
		bit valid = false;
		if ( subdir.length>0 
			&& subdir[0]!=0 
			&& ( showHidden || subdir[0]!='.')
			&& subdir!="." 
			&& subdir!=".." )
		{
			if ( dir.length == 0 )
			{
				valid = (subdir == "doc")
						|| (subdir == "src")
						|| (subdir == "images")
						|| (subdir == "screenshots")
						|| (subdir == "screenShots")
						;
			}
			else
			{
				valid = (count(dir,"doc")>0)
						|| (count(dir,"src")>0)
						|| (count(dir,"images")>0)
						|| (subdir == "screenshots")
						|| (count(dir,"screenShots")>0)
						;
			}
		}
		return valid;
	}
	
	/**
	 * loads a directory and all it's subdirectories into the tree
	 */
	private void loadDir(TreeIter parentIter, char[] dirPath, char[] superDir, bit showHidden)
	{
		static int [3] cols = [0,1,2];
		if ( dirPath.length == 0 ) return;
		try
		{
			Directory dir = new Directory(dirPath);
			foreach(char[]subdir ; dir.getDirNames().sort)
			{
				if ( validDir(superDir,subdir, showHidden) )
				{
					TreeIter iterChild = projectTreeStore.append(parentIter);
					char [][] subVals;
					subVals ~= superDir~subdir;
					subVals ~= RowType.DIRECTORY;
					subVals ~= dirPath;
					projectTreeStore.set(iterChild,cols,subVals);
					char[] subPath = dirPath~"/"~subdir;
					//printf("loadDir subDir (%d) %d %c = >%.*s<\n",subdir.length,subdir[0],subdir[0],subdir);
					//printf("loadDir subPath = >%.*s<\n",subPath);
					loadDir(parentIter, subPath, superDir~subdir~"/", showHidden);
					loadFiles(iterChild, subPath, showHidden);
					//Directory files = new Directory(subPath);
					//foreach(char[]subfile ; files.getFileNames().sort)
					//{
					//	if ( subfile.length>0 
					//		&& subfile[0]!=0 
					//		&& ( showHidden || subfile[0]!='.')
					//		&& subfile!="." 
					//		&& subfile!=".." )
					//	{
					//		TreeIter iterFile = projectTreeStore.append(iterChild);
					//		char [][] subVals;
					//		subVals  ~= subfile;
					//		subVals  ~= RowType.FILE;
					//		subVals  ~= subPath~"/"~subfile;
					//		projectTreeStore.set(iterFile,cols,subVals);
					//	}
					//}
				}
			}
		}
		catch ( DirectoryError e)
		{
			printf("ProjectView load directory error : %.*s\n",e.toString());
		}
	}

	private void loadFiles(TreeIter iterChild, char[] subPath, bit showHidden)
	{
		static int [3] cols = [0,1,2];
		Directory files = new Directory(subPath);
		foreach(char[]subfile ; files.getFileNames().sort)
		{
			if ( subfile.length>0 
				&& subfile[0]!=0 
				&& ( showHidden || subfile[0]!='.')
				&& subfile!="." 
				&& subfile!=".." )
			{
				TreeIter iterFile = projectTreeStore.append(iterChild);
				char [][] subVals;
				subVals  ~= subfile;
				subVals  ~= RowType.FILE;
				subVals  ~= subPath~"/"~subfile;
				projectTreeStore.set(iterFile,cols,subVals);
			}
		}
	}

	public void refreshProject(Project project)
	{
		if ( project !== null )
		{
			TreeIter oldIter = treeView.getProjectIter(project);
			if ( oldIter !== null )
			{
				TreeIter iter = new TreeIter();
				projectTreeStore.insertAfter(iter, null, oldIter);
	
				static int [3] cols = [0,1,2];
				char [][] vals;
				vals ~= project.getName();
				vals ~= RowType.PROJECT;
				vals ~= project.getBaseDir();
				
				projectTreeStore.set(iter,cols,vals);
		
				loadDir(iter, project.getBaseDir(), "", project.getShowHiddenFiles());
				projectTreeStore.remove(oldIter);
			}
		}
		
	}
	
	public bit removeProject(Project project, bit confirm)
	{
		bit removed = false;
		if ( project !== null )
		{
			int response = Response.NO;
			if ( confirm )
			{
				MessageDialog d = new MessageDialog(workspace, 0,
					DialogType.WARNING,DialogButtons.YES_NO,
					"Remove project " ~ project.getName()
					~ "\n(this will not delete the project contents)"
					);
				response = d.run();
				d.destroy();
			}
			if ( !confirm || (response == Response.YES) )
			{
				projects[project.getName()] = null;
				removed = true;
				//printf("going to Directory.remove()\n");
				Directory.removeFile(Leds.leds().getProjectFileName(project.getName()));
				//printf("done to Directory.remove()\n");
			}
		}
		return removed;
	}
	/**
	 * Creates the simple toolbar for the project view.
	 * @return a Toolbar
	 */
	Toolbar getToolbar()
	{	
		Dispatcher dispatcher = Dispatcher.getDispatcher();
		Toolbar toolbar = new Toolbar();
		//toolbar.setIconSize(IconSize.SMALL_TOOLBAR);
		toolbar.setIconSize(IconSize.MENU);
		toolbar.setStyle(ToolbarStyle.ICONS);
		
		// this exist because of compiler
		// problems on "forward references to classes"
		void addButton(Stock stockID, char[] tooltip, char[] action)
		{
			Button button = (Button)toolbar.insertStock(stockID,tooltip,"",-1);
			dispatcher.addButtonClickedListener(this,button,action);
		}
			
		addButton(Stock.NEW,"New","toolbar.new");
		addButton(Stock.ADD,"Add","toolbar.add");
		addButton(Stock.REMOVE,"Remove","toolbar.remove");
		
		return toolbar;
	}

	/**
	 * listens to user button clicks
	 * @param button
	 * @param action string
	 */
	void buttonClickedCallback(Button button, char [] action)
	{
		switch(action)
		{
	
			case "toolbar.new": new Project(workspace,""); break;
			//case "toolbar.add": addFile(); break;
			//case "toolbar.remove": remove(); break;
				
			default:
				printf("Project view - You press %.*s\n",action);
				break;
		}
	}

//	void addFile()
//	{
//		TreeSelection ts = treeView.getSelection();
//		TreeModel model = treeView.getModel();
//		TreePath[] paths = ts.getSelectedRows(model);
//		for ( int i = 0 ; i<paths.length ; i++ )
//		{
//			printf("path %.*s\n",paths[i].toString());
//		}
//		if ( paths.length>0 )
//		{
//			TreePath p = paths[paths.length-1];
//			p.next();
//			ts.selectPath(p);
//		}		
//	}


	Project getCurrentProject()
	{
		Project project = null;
		char[] projectName = treeView.getCurrentProjectName();
		if ( (projectName!==null) && (projectName.length>0) )
		{
			if ( projectName in projects )
			{
				project = projects[projectName];
			}
		}
		return project;
	}

	void runTool(Project project, char[] toolName, bit newConsole)
	{
		if ( project !== null )
		{
			project.runTool(toolName, newConsole);
		}
	}
	
	/**
	 * Adds the next tree entry to the selection
	 */
	void selectNext()
	{
		TreeSelection ts = treeView.getSelection();
		TreeModel model = treeView.getModel();
		TreePath[] paths = ts.getSelectedRows(model);
		for ( int i = 0 ; i<paths.length ; i++ )
		{
			printf("path %.*s\n",paths[i].toString());
		}
		if ( paths.length>0 )
		{
			TreePath p = paths[paths.length-1];
			p.next();
			ts.selectPath(p);
		}		
	}

	void selectPrev()
	{
		TreeSelection ts = treeView.getSelection();
		TreeModel model = treeView.getModel();
		TreePath[] paths = ts.getSelectedRows(model);
		if ( paths.length>0 )
		{
			TreePath p = paths[0];
			p.prev();
			ts.selectPath(p);
		}		
		
	}
	
	void newProjectFolder()
	{
		
	}
	
	TreeView getTreeView()
	{
		return treeView;
	}

	/**
	 * A TreeStore for a project
	 */
	class ProjectTreeStore : TreeStore
	{
		this(GType[] columns)
		{
			super(columns);
		}
	}
	
	ProjectTreeView createProjectTree()
	{
		projectTreeStore = new ProjectTreeStore(columns);
		//printf("projectTreeStore %X\n",projectTreeStore);
		ProjectTreeView treeView = new ProjectTreeView(projectTreeStore,this);
		TreeSelection ts = treeView.getSelection();
		ts.setMode(SelectionMode.MULTIPLE);
		treeView.setHeadersVisible(false);
		TreeViewColumn column = new TreeViewColumn("Elements",new CellRendererText(),"text", 0);
		treeView.appendColumn(column);

		// hidden column
		//column = new TreeViewColumn("CODE",new CellRendererText(),"text", 1);
		//treeView.appendColumn(column);
		// hidden column
		//column = new TreeViewColumn("VALUE",new CellRendererText(),"text", 2);
		//treeView.appendColumn(column);
		//printf("projectTreeStore exit %X\n",projectTreeStore);

		return treeView;
	}

	/**
	 * A TreeView to display a project structure
	 */
	class ProjectTreeView : TreeView , MouseButtonListener , MenuItemListener
	{
	
		Menu projectPopupMenu;
		Menu directoryPopupMenu;
		Menu filePopupMenu;
		
		ProjectView projectView;

		TreeStore treeStore;
		
		this(TreeStore treeStore, ProjectView projectView)
		{
			super(treeStore);
			this.treeStore = treeStore;
			this.projectView = projectView;
		//printf("treeStore %X\n",this.treeStore);
			Dispatcher d = Dispatcher.getDispatcher();
			projectPopupMenu = buildProjectMenu();
			directoryPopupMenu = buildDirectoryMenu();
			filePopupMenu = buildFileMenu();
			d.addMouseButtonListener(this,this);
			//d.addMenuItemListener(this,this);
		}

		public TreeIter getProjectIter(Project project)
		{
			TreeIter iter = new TreeIter();
			bit got = treeStore.getIterFirst(iter);
			while ( got && (iter.getValueString(0) != project.getName() ) )
			{
				got = treeStore.iterNext(iter);
			}
			
			return got ? iter : null;
		}
	
		
		Menu buildProjectMenu()
		{
			Menu menu = new Menu();
			MenuItem item;
			
			menu.append(new MenuItem(this,"_Build","projectPopupMenu.build"));
			menu.append(new MenuItem(this,"_Run","projectPopupMenu.run"));
			menu.append(new SeparatorMenuItem());
			menu.append(new MenuItem(this,"Refresh","projectPopupMenu.refresh"));
			menu.append(new MenuItem(this,"User Documentation tool","projectPopupMenu.runUserDoc"));
			menu.append(new MenuItem(this,"Developer Documentation tool","projectPopupMenu.runDevelDoc"));
			menu.append(new MenuItem(this,"Private Documentation tool","projectPopupMenu.runPrivateDoc"));
			menu.append(new SeparatorMenuItem());
			menu.append(new MenuItem(this,"Remove","projectPopupMenu.remove"));
			menu.append(new SeparatorMenuItem());
			menu.append(new MenuItem(this,"_Properties","projectPopupMenu.properties"));

			return menu;
		}
		
		Menu buildDirectoryMenu()
		{
			Menu menu = new Menu();
			MenuItem item;
			
			menu.append(new MenuItem(this,"_Filter","directoryPopupMenu.filter"));
			menu.append(new MenuItem(this,"_Hide","directoryPopupMenu.hide"));

			return menu;
		}
		
		Menu buildFileMenu()
		{
			Menu menu = new Menu();
			MenuItem item;
			
			menu.append(new MenuItem(this,"_Open","filePopupMenu.open"));
			menu.append(new MenuItem(this,"_Copy","filePopupMenu.copy"));
			menu.append(new MenuItem(this,"_Paste","filePopupMenu.paste"));
			menu.append(new MenuItem(this,"_Cut","filePopupMenu.cut"));
			menu.append(new MenuItem(this,"_Remove","filePopupMenu.remove"));

			return menu;
		}

		char[] getCurrentProjectName()
		{
			char[] projectName = null;

			TreeIter[] iters = getSelectedIters();
			if ( iters.length > 0 )
			{
				TreeIter iter = iters[0].getGrandParent();
				if ( iter !== null )
				{
					projectName = iter.getValueString(0).dup;
				}
			}

			return projectName;
		}

		
		bit mouseButtonPressCallback(Widget widget, EventButton eventButton)
		{
			//printf("mouse pressed on project tree view with button %d and time %d\n",eventButton.button(),eventButton.getTime());
			if ( eventButton.button() == 3 )
			{
			}
			else
			{
			}
			return false;   // false : the event is passed to other listeners
		}
		
		bit mouseButtonReleaseCallback(Widget widget, EventButton eventButton)
		{
			//printf("mouse release on project tree view with button %d\n",eventButton.button());
			if ( eventButton.button() == 3 )
			{
				//printf("going for the menusssssssssssssss\n");
				TreeIter iter = getSelectedIter();
				//printf("iter = %X\n",iter);
				if ( iter !== null )
				{
					switch(iter.getValueString(1))
					{
						case RowType.PROJECT:  
							projectPopupMenu.popup(eventButton.button(),eventButton.getTime());
							break;
							
						case RowType.DIRECTORY:  
							directoryPopupMenu.popup(eventButton.button(),eventButton.getTime());
							break;
							
						case RowType.FILE:  
							filePopupMenu.popup(eventButton.button(),eventButton.getTime());
							break;
						
						default:
							break;
					}
				}
				else
				{
					//projectMenu.popup(eventButton.button(),eventButton.getTime());
				}
			}
			else
			{
			}
			return false;   // false : the event is passed to other listeners
		}

		/**
		 * Menu listener callback
		 */
		void activateCallback(MenuItem menuItem, char [] action)
		{
	//printf("activateCallback 1\n");
			TreeIter[] iters = getSelectedIters();
	//printf("activateCallback 2\n");
			char[] projectName = iters[0].getValueString(0);
			Project project = null;
			if ( projectName in projectView.projects)
			{
				project = projectView.projects[projectName];
			}
			switch(action)
			{
				
				case "projectPopupMenu.build":
					projectView.runTool(project,"makeCommand:",true);
					break;
					
				case "projectPopupMenu.run":			
					projectView.runTool(project,"runCommand:",true);
					break;
					
				case "projectPopupMenu.refresh":
					projectView.refreshProject(project);
					break;
					
				case "projectPopupMenu.runUserDoc":
					projectView.runTool(project,"userDocCommand:",true);
					break;

				case "projectPopupMenu.runDevelDoc":
					projectView.runTool(project,"develDocCommand:",true);
					break;

				case "projectPopupMenu.runPrivateDoc":
					projectView.runTool(project,"privateDocCommand:",true);
					break;

				case "projectPopupMenu.remove":
					if ( projectView.removeProject(project,true) )
					{
						printf("-------- remove iters[0] ----------\n");
						treeStore.remove(iters[0]);
					}
					break;

				case "projectPopupMenu.properties":
					if ( project !== null )
					{
						project.guiSetup(false);
					}
					break;
					
				case "filePopupMenu.open":
		//printf("filePopupMenu.open 1\n");
					foreach ( TreeIter iter ; iters )
					{
		//printf("filePopupMenu.open 2\n");
						if ( iter !== null && "F"==iter.getValueString(1) )
						{
		//printf("filePopupMenu.open 3\n");
							printf("projectView %X\n",projectView);
		//printf("filePopupMenu.open 4\n");
							printf("projectView.workspace %X\n",projectView.workspace);
		//printf("filePopupMenu.open 5\n");
							projectView.workspace.openFile(iter.getValueString(2));
						}
		//printf("filePopupMenu.open 6\n");
					}
		//printf("filePopupMenu.open 7\n");
					break;
					
				default:
					printf("Menu option %.*s\n", action);
					break;
			}
		}
		
		void activateItemCallback(MenuItem menuItem, char [] action)
		{
			switch(action)
			{
				default:
					printf("Menu option %.*s\n", action);
					break;
			}
		}
		
	}
	
}